# -*- coding: utf-8 -*-
import sys
PY3 = sys.version_info >= (3,)
if PY3:
    from io import StringIO, BytesIO as StreamIO
else:
    from StringIO import StringIO, StringIO as StreamIO
import os
from os.path import abspath, dirname
from shutil import rmtree
import tempfile
from unittest import TestCase

SKOOLKIT_HOME = abspath(dirname(dirname(__file__)))
sys.path.insert(0, SKOOLKIT_HOME)
from skoolkit import bin2tap, skool2asm, skool2ctl, skool2html, skool2sft, sna2skool

class Stream:
    def __init__(self):
        self.buffer = StringIO()

    def write(self, text):
        self.buffer.write(text)

    def writelines(self, lines):
        self.buffer.writelines(lines)

    def flush(self):
        return

    def getvalue(self):
        return self.buffer.getvalue()

    def clear(self):
        self.buffer.seek(0)
        self.buffer.truncate()

class SkoolKitTestCase(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.stdin = sys.stdin
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        sys.stdout = self.out = Stream()
        sys.stderr = self.err = Stream()
        self.tempfiles = []
        self.tempdirs = []

    def tearDown(self):
        self.remove_files()
        sys.stdin.close()
        sys.stdin = self.stdin
        sys.stdout = self.stdout
        sys.stderr = self.stderr

    def remove_files(self):
        for f in self.tempfiles:
            if os.path.isfile(f):
                os.remove(f)
        self.tempfiles = []
        for d in self.tempdirs:
            if os.path.isdir(d):
                rmtree(d, True)
        self.tempdirs = []

    def clear_streams(self):
        self.out.clear()
        self.err.clear()

    def make_directory(self, path):
        if path and not os.path.isdir(path):
            os.makedirs(path)
            dirs = [path]
            while True:
                path = dirname(path)
                if path in ('', '.'):
                    break
                dirs.append(path)
            self.tempdirs.append(dirs[-1])

    def write_text_file(self, contents='', path=None, suffix=''):
        if path is None:
            fd, path = tempfile.mkstemp(suffix=suffix, text=True)
            path = path.replace(os.path.sep, '/')
            f = os.fdopen(fd, 'wt')
        else:
            self.make_directory(dirname(path))
            f = open(path, 'wt')
        self.tempfiles.append(path)
        f.write(contents)
        f.close()
        return path

    def write_bin_file(self, data=(), path=None, suffix=''):
        if path is None:
            fd, path = tempfile.mkstemp(suffix=suffix)
            path = path.replace(os.path.sep, '/')
            f = os.fdopen(fd, 'wb')
        else:
            self.make_directory(dirname(path))
            f = open(path, 'wb')
        self.tempfiles.append(path)
        f.write(bytearray(data))
        f.close()
        return path

    def write_stdin(self, contents):
        sys.stdin = open(self.write_text_file(contents), 'rt')

    def assert_output_equal(self, output, expected, prefixes=False):
        for i, line in enumerate(expected):
            if i < len(output):
                if prefixes:
                    self.assertEqual(output[i][:len(line)], line)
                else:
                    self.assertEqual(output[i], line)
        if len(output) > len(expected):
            self.fail("Unexpected extra line in output: '%s'" % output[len(expected)])
        elif len(expected) > len(output):
            self.fail("Expected lines missing from output, starting with: '%s'" % expected[len(output)])

    def to_lines(self, text, strip_cr):
        # Use rstrip() to remove '\r' characters (useful on Windows)
        if strip_cr:
            lines = [line.rstrip() for line in text.split('\n')]
        else:
            lines = text.split('\n')
        if lines[-1] == '':
            lines.pop()
        return lines

    def _run_skoolkit_command(self, cmd, args, out_lines, err_lines, strip_cr):
        self.clear_streams()
        try:
            cmd(args.split())
        except SystemExit:
            pass
        out = self.to_lines(self.out.getvalue(), strip_cr) if out_lines else self.out.getvalue()
        err = self.to_lines(self.err.getvalue(), strip_cr) if err_lines else self.err.getvalue()
        return out, err

    def run_bin2tap(self, args='', out_lines=True, err_lines=False, strip_cr=True):
        return self._run_skoolkit_command(bin2tap.main, args, out_lines, err_lines, strip_cr)

    def run_skool2asm(self, args='', out_lines=True, err_lines=False, strip_cr=True):
        return self._run_skoolkit_command(skool2asm.main, args, out_lines, err_lines, strip_cr)

    def run_skool2ctl(self, args='', out_lines=True, err_lines=False, strip_cr=True):
        return self._run_skoolkit_command(skool2ctl.main, args, out_lines, err_lines, strip_cr)

    def run_skool2html(self, args='', out_lines=True, err_lines=False, strip_cr=True):
        return self._run_skoolkit_command(skool2html.main, args, out_lines, err_lines, strip_cr)

    def run_skool2sft(self, args='', out_lines=True, err_lines=False, strip_cr=True):
        return self._run_skoolkit_command(skool2sft.main, args, out_lines, err_lines, strip_cr)

    def run_sna2skool(self, args='', out_lines=True, err_lines=False, strip_cr=True):
        return self._run_skoolkit_command(sna2skool.main, args, out_lines, err_lines, strip_cr)
