# -*- coding: utf-8 -*-
import unittest

from skoolkittest import SkoolKitTestCase
from skoolkit import write_text, SkoolKitError

TEST_NO_OPTIONS_SKOOL = """; @start
; @org=24593

; Let's test some @ofix directives
c24593 NOP
; @ofix=LD A,C
 24594 LD A,B
; @ofix-begin
 24595 LD B,A
; @ofix+else
 24595 LD B,C
; @ofix+end

; Let's test some @bfix directives
c24596 NOP
; @bfix=LD C,B
 24597 LD C,A
; @bfix-begin
 24598 LD D,A
; @bfix+else
 24598 LD D,B
; @bfix+end

; Let's test the @rfix block directive
c24599 NOP
; @rfix-begin
 24600 LD E,A
; @rfix+else
 24600 LD E,B
; @rfix+end

; Let's test the @ssub directive
; @ssub=JP (HL)
c24601 RET

; Let's test the @rsub block directive
c24602 NOP
; @rsub-begin
 24603 LD A,0
; @rsub+else
 24603 XOR A
; @rsub+end
"""

TEST_d_SKOOL = """; @start
; Begin
c$8000 RET
"""

TEST_t_SKOOL = """; @start
; @org=24576

; Routine
c24576 RET
"""

TEST_l_SKOOL = """; @start
; @org=24576

; Routine
; @label=DOSTUFF
c24576 NOP
"""

TEST_i_SKOOL = """; @start
; Do nothing
c50000 RET ; Return
"""

TEST_f_SKOOL = """; @start
; Let's test some @ofix directives
c24593 NOP
; @ofix=LD A,C
 24594 LD A,B
; @ofix-begin
 24595 LD B,A
; @ofix+else
 24595 LD B,C
; @ofix+end

; Let's test some @bfix directives
c24596 NOP
; @bfix=LD C,B
 24597 LD C,A
; @bfix-begin
 24598 LD D,A
; @bfix+else
 24598 LD D,B
; @bfix+end

; Let's test the @rfix block directive
c24599 NOP
; @rfix-begin
 24600 LD E,A
; @rfix+else
 24600 LD E,B
; @rfix+end
"""

TEST_s_SKOOL = """; @start
; Let's test the @ssub directive
; @ssub=JP (HL)
c24601 RET
"""

TEST_r_SKOOL = """; @start
; Let's test some @ofix directives
c24593 NOP
; @ofix=LD A,C
 24594 LD A,B
; @ofix-begin
 24595 LD B,A
; @ofix+else
 24595 LD B,C
; @ofix+end

; Let's test some @bfix directives
c24596 NOP
; @bfix=LD C,B
 24597 LD C,A
; @bfix-begin
 24598 LD D,A
; @bfix+else
 24598 LD D,B
; @bfix+end

; Let's test the @rfix block directive
c24599 NOP
; @rfix-begin
 24600 LD E,A
; @rfix+else
 24600 LD E,B
; @rfix+end

; Let's test the @ssub directive
; @ssub=JP (HL)
c24601 RET

; Let's test the @rsub block directive
c24602 NOP
; @rsub-begin
 24603 LD A,0
; @rsub+else
 24603 XOR A
; @rsub+end
"""

TEST_Q_SKOOL = """; @start
; Do nothing
c30000 RET
"""

TEST_W_SKOOL = """; This skool file generates warnings

; @start
; Routine at 24576
;
; Used by the routine at 24576.
c24576 JP 24576
"""

TEST_U_SKOOL = """; @start
; Start the game
; @label=start
c49152 nop
"""

TEST_D_SKOOL = """; @start
; Begin
c$8000 JP $ABCD
"""

TEST_H_SKOOL = """; @start
; Begin
c$8000 JP 56506
"""

TEST_C_SKOOL = """; @start
; Begin
c32768 JR 32770

; End
c32770 JR 32768
"""

TEST_NO_HTML_ESCAPE_SKOOL = """; @start
; Text
t24576 DEFM "&<>" ; a <= b & b >= c
"""

TEST_MACRO_EXPANSION_SKOOL = """; @start
; Data
b$6003 DEFB 123 ; #REGa=0
 $6004 DEFB $23 ; '#'
"""

TEST_WRITER_SKOOL = """; @start
; @writer={0}
; Begin
c24576 RET
"""

USAGE = """Usage: skool2asm.py [options] FILE

  Convert a skool file into an ASM file, written to standard output. FILE may
  be a regular file, or '-' for standard input.

Options:
  -q    Be quiet
  -w    Suppress warnings
  -d    Use CR+LF to end lines
  -t    Use tab to indent instructions (default indentation is 2 spaces)
  -l    Write disassembly in lower case
  -u    Write disassembly in upper case
  -D    Write disassembly in decimal
  -H    Write disassembly in hexadecimal
  -i N  Set instruction field width to N (default=23)
  -f N  Apply fixes:
          N=0: None (default)
          N=1: @ofix only
          N=2: @ofix and @bfix
          N=3: @ofix, @bfix and @rfix (implies -r)
  -c    Create default labels for unlabelled instructions
  -s    Use safe substitutions (@ssub)
  -r    Use relocatability substitutions too (@rsub) (implies '-f 1')"""

class TestAsmWriter:
    def __init__(self, *args):
        pass

    def write(self):
        write_text('OK')

class Skool2AsmTest(SkoolKitTestCase):
    def test_no_arguments(self):
        output, error = self.run_skool2asm(err_lines=True)
        self.assert_output_equal(error, USAGE.split('\n'))

    def test_invalid_option(self):
        output, error = self.run_skool2asm('-x', err_lines=True)
        self.assert_output_equal(error, USAGE.split('\n'))

    def test_no_options(self):
        asm = self.get_asm(skool=TEST_NO_OPTIONS_SKOOL)
        self.assertEqual(asm[0], '  ORG 24593') # no crlf, tabs or lower case
        self.assertEqual(asm[4], '  LD A,B')    # No @ofix
        self.assertEqual(asm[5], '  LD B,A')    # @ofix-
        self.assertEqual(asm[9], '  LD C,A')    # No @bfix
        self.assertEqual(asm[10], '  LD D,A')   # @bfix-
        self.assertEqual(asm[14], '  LD E,A')   # @rfix-
        self.assertEqual(asm[17], '  RET')      # No @ssub
        self.assertEqual(asm[21], '  LD A,0')   # @rsub-

    def test_q(self):
        skoolfile = self.write_text_file(TEST_Q_SKOOL, suffix='.skool')
        output, error = self.run_skool2asm('-q {0}'.format(skoolfile))
        self.assertEqual(error, '')

    def test_w(self):
        skoolfile = self.write_text_file(TEST_W_SKOOL, suffix='.skool')
        output, error = self.run_skool2asm('-q -w {0}'.format(skoolfile))
        self.assertEqual(error, '')

    def test_d(self):
        asm = self.get_asm('-d', TEST_d_SKOOL, strip_cr=False)
        self.assertEqual(asm[0][-1], '\r')

    def test_t(self):
        asm = self.get_asm('-t', TEST_t_SKOOL)
        self.assertEqual(asm[0], '\tORG 24576')
        self.assertEqual(asm[3], '\tRET')

    def test_l(self):
        asm = self.get_asm('-l', TEST_l_SKOOL)
        self.assertEqual(asm[0], '  org 24576')
        self.assertEqual(asm[3], 'DOSTUFF:') # Labels unaffected
        self.assertEqual(asm[4], '  nop')

    def test_u(self):
        asm = self.get_asm('-u', TEST_U_SKOOL)
        self.assertEqual(asm[1], 'start:') # Labels unaffected
        self.assertEqual(asm[2], '  NOP')

    def test_D(self):
        asm = self.get_asm('-D', TEST_D_SKOOL)
        self.assertEqual(asm[1], '  JP 43981')

    def test_H(self):
        asm = self.get_asm('-H', TEST_H_SKOOL)
        self.assertEqual(asm[1], '  JP $DCBA')

    def test_i(self):
        width = 30
        asm = self.get_asm('-i {0}'.format(width), TEST_i_SKOOL)
        self.assertEqual(asm[1].find(';'), width + 3)

    def test_f0(self):
        asm = self.get_asm('-f 0', TEST_f_SKOOL)
        self.assertEqual(asm[2], '  LD A,B')
        self.assertEqual(asm[3], '  LD B,A')
        self.assertEqual(asm[7], '  LD C,A')
        self.assertEqual(asm[8], '  LD D,A')
        self.assertEqual(asm[12], '  LD E,A')

    def test_f1(self):
        asm = self.get_asm('-f 1', TEST_f_SKOOL)
        self.assertEqual(asm[2], '  LD A,C')
        self.assertEqual(asm[3], '  LD B,C')
        self.assertEqual(asm[7], '  LD C,A')
        self.assertEqual(asm[8], '  LD D,A')
        self.assertEqual(asm[12], '  LD E,A')

    def test_f2(self):
        asm = self.get_asm('-f 2', TEST_f_SKOOL)
        self.assertEqual(asm[2], '  LD A,C')
        self.assertEqual(asm[3], '  LD B,C')
        self.assertEqual(asm[7], '  LD C,B')
        self.assertEqual(asm[8], '  LD D,B')
        self.assertEqual(asm[12], '  LD E,A')

    def test_f3(self):
        asm = self.get_asm('-f 3', TEST_f_SKOOL)
        self.assertEqual(asm[2], '  LD A,C')
        self.assertEqual(asm[3], '  LD B,C')
        self.assertEqual(asm[7], '  LD C,B')
        self.assertEqual(asm[8], '  LD D,B')
        self.assertEqual(asm[12], '  LD E,B')

    def test_s(self):
        asm = self.get_asm('-s', TEST_s_SKOOL)
        self.assertEqual(asm[1], '  JP (HL)')

    def test_r(self):
        asm = self.get_asm('-r', TEST_r_SKOOL)
        self.assertEqual(asm[2], '  LD A,C')   # @ofix
        self.assertEqual(asm[3], '  LD B,C')   # @ofix+
        self.assertEqual(asm[7], '  LD C,A')   # No @bfix
        self.assertEqual(asm[8], '  LD D,A')   # @bfix-
        self.assertEqual(asm[12], '  LD E,A')  # @rfix-
        self.assertEqual(asm[15], '  JP (HL)') # @ssub
        self.assertEqual(asm[19], '  XOR A')   # @rsub+

    def test_c(self):
        asm = self.get_asm('-c', TEST_C_SKOOL)
        self.assertEqual(asm[1], 'L32768:')
        self.assertEqual(asm[2], '  JR L32770')
        self.assertEqual(asm[5], 'L32770:')
        self.assertEqual(asm[6], '  JR L32768')

    def test_writer(self):
        # Test a writer with no module or package name
        writer = 'AbsoluteAsmWriter'
        skoolfile = self.write_text_file(TEST_WRITER_SKOOL.format(writer), suffix='.skool')
        try:
            self.run_skool2asm(skoolfile)
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], "Invalid class name: '{0}'".format(writer))

        # Test a writer in a nonexistent module
        writer = 'nonexistentmodule.AsmWriter'
        skoolfile = self.write_text_file(TEST_WRITER_SKOOL.format(writer), suffix='.skool')
        try:
            self.run_skool2asm(skoolfile)
            self.fail()
        except SkoolKitError as e:
            err_t = "Failed to import class nonexistentmodule.AsmWriter: No module named {0}nonexistentmodule{0}"
            self.assertTrue(e.args[0] in (err_t.format(''), err_t.format("'")))

        # Test a writer that doesn't exist
        writer = 'test_skool2asm.NonexistentAsmWriter'
        skoolfile = self.write_text_file(TEST_WRITER_SKOOL.format(writer), suffix='.skool')
        try:
            self.run_skool2asm(skoolfile)
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], "No class named 'NonexistentAsmWriter' in module 'test_skool2asm'")

        # Test a writer that exists
        asm = self.get_asm(skool=TEST_WRITER_SKOOL.format('test_skool2asm.TestAsmWriter'))
        self.assertEqual(asm[0], 'OK')

    def test_no_html_escape(self):
        asm = self.get_asm(skool=TEST_NO_HTML_ESCAPE_SKOOL)
        self.assertEqual(asm[1], '  DEFM "&<>"              ; a <= b & b >= c')

    def test_macro_expansion(self):
        asm = self.get_asm(skool=TEST_MACRO_EXPANSION_SKOOL)
        self.assertEqual(asm[1], '  DEFB 123                ; A=0')
        self.assertEqual(asm[2], "  DEFB $23                ; '#'")

    def get_asm(self, args='', skool='', out_lines=True, err_lines=True, strip_cr=True):
        skoolfile = self.write_text_file(skool, suffix='.skool')
        output, error = self.run_skool2asm('{0} {1}'.format(args, skoolfile), out_lines, err_lines, strip_cr)
        self.assertEqual('Wrote ASM to stdout', error[-1][:19], 'Error(s) while running skool2asm.main() with args "{0}"'.format(args))
        return output

if __name__ == '__main__':
    unittest.main()
